/***********************************************************************\
 * Wzorzec projektowy puli.                Konrad Grochowski, nr 192621 *
 *                                                                      *
 * Zastosowania:                                                        *
 *           - Zarzdzanie grup obiektw reprezentujcych              *
 *             zasoby wielokrotnego uycia                              *
 *           - Ograniczenie kosztw tworzenia i usuwania obiektw       *
 *                                                                      *
 * Opis:                                                                *
 *        Wzorzec realizowany jest zgodnie ze schematem przedstawionym  *
 *        poniej. Obiekt Klient wykorzystuje obiekt PulaObiektow jako  *
 *        fabryk dla ObiektWielokrotnegoUzytku a po zakoczeniu        *
 *        interakcji z uzyskanym obiektem zwraca go do puli, gdzie jest *
 *        on przywracany do stanu "czystego" i moze nadawa do          *
 *        ponownego uytku przez innego klienta puli. Dziki temu       *
 *        zminimalizowany jest czas powicony na alokacj zasobw, co  *
 *        moe mie due znaczenie np. w przypadku zasobw systemowych  *
 *        lub bardzo duych obiektw. Uytkownik jest take zwolniony   *
 *        z odpowiedzialnoci za poprawne zakoczenie ycia obiektu -   *
 *        odpowiada za to pula.                                         *
 *        Zachowanie puli moe by dostosowane do specyficznych         *
 *        zastosowa. Moe mie stay lub zmienny (ograniczony od gry) *
 *        rozmiar, moe tworzy obiekty gdy jest to konieczne, lub      *
 *        zaalokowa wszystkie od razu.                                 *
 *                                                                      *
 *     /--------\ 0...n                                                 *
 *     | Klient |---------------------------------------                *
 *     \--------/                                       |               *
 *         |                                           \|/ 1            *
 *                                             /------------------\     *
 *         |                                   |   PulaObiektow   |     *
 *                                             |------------------|     *
 *         |                                   |  pobierzObiekt() |     *
 *                                             |  zwolnijObiekt() |     *
 *         |                                   \------------------/     *
 *                                                      /\              *
 *         |                                            \/ 1            *
 *                                                      |               *
 *        \|/                                           |               *
 *     /---------------------------\                    |               *
 *     | ObiektWielokrotnegoUzytku |                    |               *
 *     |---------------------------|/___________________|               *
 *     |        wyczysc()          |\ 0...n                             *
 *     \---------------------------/                                    *
 *                                                                      *
 \**********************************************************************/
#include <cassert>
#include <vector>
#include <iostream>
/**
 * Obiekt ktory bedzie skladnikiem puli.
 * W tym przykadzie jest to obiekt reprezentujcy
 * wiadomo jakiego protokou, wymagajc duego buforu
 * na dane.
 */
class Message
{
public:
  // Ten rozmiar powoduje ze nie opaca si
  // realokowa obiektw tego typu za kadym razem gdy s potrzebne
  static const unsigned int MSG_SIZE = 100000000;

  Message() { reset(); }
  virtual ~Message() {}

  // Tutaj s funkcje uytkowe
  int getUserId() const { return m_userId; }
  void setUserId(int userId_) { m_userId = userId_; }
  const char* getMessageBuffer() const { return m_messageBuffer; }
  char* getMessageBuffer() { return m_messageBuffer; }

  // A to funkcja uywana przez pul do "wyczyszczenia" stanu obiektu
  void reset();

private:
  int m_userId;
  char m_messageBuffer[MSG_SIZE];
};

/**
 * Klasa reprezentujca sam pul.
 * Tworzy obiekty od razu w konstruktorze - potem szybciej mona
 * przetwarza wiadomoci. Mona da dodatkowy parametr pozwalajcy
 * na powikszanie si puli do pewnego limitu (tzw. "wskanik wysokiej wody"
 * ang. "high water mark")
 */
class MessagePool
{
public:
  static const unsigned int FIXED_SIZE = 0;

  /**
   * Tworzy pul o podanym rozmiarze i ewentualnym rozmiarze max.
   */
  MessagePool(unsigned int initialSize_,
              unsigned int maxSize_ = FIXED_SIZE);
  virtual ~MessagePool();

  /* Zwraca jeden z elementw puli, lub NULL jeli pula nie ma wicej elementw
  * O ile to moliwe powiksza pul o nowy obiekt (jeli pula mia podany
  * drugi parametr w konstruktorze
  */
  Message* takeMessage();
  /** Przyjmuje obiekt z powrotem do puli i czyci jego stan.
   * Kady obiekt pobrany musi by zwrcony do niej t metod
   */
  void releaseMessage(Message* msg);

private:
  typedef std::vector<Message*> MessageVector;
  unsigned int m_maxSize, m_actualSize;
  MessageVector m_messageVector;
};

int main(int argc, char * argv[])
{
  using namespace std;
  MessagePool pool(2); // mala pula dla przykadu
  Message* m1 = pool.takeMessage();
  Message* m2 = pool.takeMessage();
  Message* m3 = pool.takeMessage();

  // 3 wskanik jest NULLem - bo pula sie skoczya
  cout << "M1: " << m1 << " M2: " << m2 << " M3: " << m3 << endl;

  // symulacja odczytu wiadomoci z protokou np sieciowego
  memset(m1->getMessageBuffer(),'A',511);
  m2->getMessageBuffer()[512] = '\0';
  // wypisanie - "uycie" - wiadomoci
  cout << "M1 Message: " << m1->getMessageBuffer() << endl;
  // zwolnienie uytej ju wiadomoci
  pool.releaseMessage(m1);

  // pobranie obiektu na now wiadomo
  m3 = pool.takeMessage();
  // to jest ten sam obiekt co za pierwszym razem
  cout << "M1==M3: " << (m1==m3) << endl;
  // ale wyczyszczony i gotowy do ponownego uycia
  cout << "M3 Message: " << m3->getMessageBuffer() << endl;

  // czyszczenie na koniec
  pool.releaseMessage(m2);
  pool.releaseMessage(m3);

  return 0;
}

void Message::reset()
{
  m_userId = -1;
  m_messageBuffer[0] = 0;
}

MessagePool::MessagePool(unsigned int initialSize_,
                         unsigned int maxSize_)
  : m_maxSize(maxSize_ != FIXED_SIZE ? maxSize_ : initialSize_)
{
  m_messageVector.reserve(m_maxSize);
  for(unsigned int i = 0; i < initialSize_; ++i)
  {
    Message* m = new Message();
    m_messageVector.push_back(m);
  }
  m_actualSize = m_messageVector.size();
}

MessagePool::~MessagePool()
{
  // wane sprawdzenie, czy uytkownicy wszystko oddaj co wzieli
  assert(m_messageVector.size() == m_actualSize);

  for(MessageVector::iterator i = m_messageVector.begin();
      i != m_messageVector.end();
      ++i)
    delete *i;
}

Message* MessagePool::takeMessage()
{
  if(m_messageVector.size() > 0)
  {
    Message* ret = m_messageVector.back();
    m_messageVector.pop_back();
    return ret;
  }

  if(m_actualSize == m_maxSize) return NULL;

  ++m_actualSize;
  return new Message();
}

void MessagePool::releaseMessage(Message* msg)
{
  if(!msg) return;
  assert(m_actualSize >= m_messageVector.size());
  msg->reset();
  m_messageVector.push_back(msg);
}
